import warnings

from alliance_platform.frontend.bundler.resource_registry import FrontendResourceRegistry

warnings.warn(
    "alliance_platform.frontend.asset_registry is deprecated; use alliance_platform.frontend.resource_registry instead",
    DeprecationWarning,
    stacklevel=2,
)

FrontendAssetRegistry = FrontendResourceRegistry



from __future__ import annotations

from dataclasses import dataclass
import mimetypes
from pathlib import Path


def get_content_type(src: str | Path | None) -> str | None:
    if not src:
        # If no source assume js, e.g. a common chunk
        return "text/javascript"
    ext = "".join(Path(src).suffixes).lstrip(".").lower()
    if ext in ["js", "jsx", "mjs", "ts", "tsx"]:
        return "text/javascript"
    if ext in ["css", "css.ts"]:
        return "text/css"
    return mimetypes.guess_type(src)[0]


@dataclass(frozen=True)
class FrontendResource:
    """Represents a resource to be bundled by the bundler

    Previously, we just tracked files that a bundler needed to care about but this limits what you can do. For example,
    you have to bundle everything at that path, even if you use one export in the case with ES Modules. By storing the
    specifics about the resource, the bundler can then optimise this and - in the case of ES Modules - only bundle the
    resources actually used (e.g. perform tree shaking). It also means we can more easily detect what resource we
    are dealing with without relying on only inferring it from the filename.
    """

    #: The path to the resource
    path: Path

    @property
    def content_type(self) -> str | None:
        """Get the content type string for this resource"""
        return get_content_type(self.path)

    @classmethod
    def from_path(cls, path: Path | str):
        """Given a path, return the most appropriate resource for it"""
        if isinstance(path, str):
            path = Path(path)
        content_type = get_content_type(path)
        if content_type == "text/css":
            if path.suffixes == [".css", ".ts"]:
                from alliance_platform.frontend.templatetags.vanilla_extract import VanillaExtractResource

                return VanillaExtractResource(path)
            return CssResource(path)
        if content_type == "text/javascript":
            return JavascriptResource(path)
        if content_type and content_type.startswith("image/"):
            return ImageResource(path)
            # assume js as default
        return JavascriptResource(path)

    def serialize(self):
        """Serialize the resource for use by the bundler.

        Returns a dict with, at minimum, the following:

        - ``type`` - a string identifying the resource type. Currently, this will be one of "javascript", "css", "esmodule", "image".
        - ``path`` - path to the file for the resource

        Then any additional specific to the resource (e.g. the import details for esmodule usage)

        This is used by :djmanage:`extract_frontend_resources` to export the resource for use in the bundler.
        """
        raise NotImplementedError()

    def is_substitutable_for(self, other: FrontendResource) -> bool:
        """Check if this resource can be used in place of another resource."""
        return self == other


class JavascriptResource(FrontendResource):
    """A plain javascript resource.

    See also :class:`~alliance_platform.frontend.bundler.resource.ESModuleResource`"""

    @property
    def content_type(self) -> str:
        return "text/javascript"

    def serialize(self):
        return {
            "type": "javascript",
            "path": str(self.path),
        }

    def is_substitutable_for(self, other: FrontendResource) -> bool:
        """
        A JavascriptResource is a less specific version of ESModuleResource. When JavascriptResource is used the
        entire path will be bundled, so it covers off all possible exports used by ESModuleResource and so can
        be used in place of it."""
        return self == other or isinstance(other, ESModuleResource) and self.path == other.path


@dataclass(frozen=True)
class ESModuleResource(JavascriptResource):
    """An ESModule resource

    This allows the bundler to perform tree shaking rather than bundling the whole file and its dependencies.
    """

    import_name: str
    is_default_import: bool

    def serialize(self):
        return {
            "type": "esmodule",
            "path": str(self.path),
            "importName": self.import_name,
            "isDefaultImport": self.is_default_import,
        }


class CssResource(FrontendResource):
    """A CSS resource"""

    @property
    def content_type(self) -> str:
        return "text/css"

    def serialize(self):
        return {
            "type": "css",
            "path": str(self.path),
        }


class VanillaExtractResource(CssResource):
    """A Vanilla Extract CSS resource

    This will be a file with extension .css.ts
    """

    pass


class ImageResource(FrontendResource):
    """An image resource"""

    def serialize(self):
        return {
            "type": "image",
            "path": str(self.path),
        }



import json
from unittest import mock

from alliance_platform.frontend.bundler.ssr import BundlerAssetServerSideRenderer
from alliance_platform.frontend.bundler.ssr import SSRItem
from django.test import SimpleTestCase
import requests

from tests.test_utils import bundler as bundler_test_utils
from tests.test_utils import override_ap_frontend_settings


class DummySSRItem(SSRItem):
    def get_ssr_type(self):
        return "Component"

    def get_ssr_payload(self, ssr_context):
        return {
            "component": "div",
            "props": {"label": "hello"},
            "identifierPrefix": "ITEM",
        }


class MockResponse:
    def __init__(self, data, status_code=200):
        self._data = data
        self.status_code = status_code
        self.content = json.dumps(data).encode()

    def json(self):
        return self._data


class SSRRendererTests(SimpleTestCase):
    def create_bundler(self):
        return bundler_test_utils.TestViteBundler(
            **bundler_test_utils.bundler_kwargs,  # type: ignore[arg-type]
            mode="development",
        )

    @override_ap_frontend_settings()
    def test_payload_uses_v2_shape(self):
        bundler = self.create_bundler()
        renderer = BundlerAssetServerSideRenderer({"PLACEHOLDER": DummySSRItem()})

        def mocked_post(*args, **kwargs):
            payload = json.loads(kwargs["data"])
            self.assertEqual(payload["version"], 2)
            self.assertTrue(isinstance(payload["requestId"], str) and payload["requestId"])
            self.assertIn("items", payload)
            self.assertIn("requiredImports", payload)
            return MockResponse(
                {
                    "renderedItems": {
                        "PLACEHOLDER": {
                            "html": "<div>rendered</div>",
                            "renderErrors": [],
                        }
                    },
                    "errors": {},
                }
            )

        with mock.patch("alliance_platform.frontend.bundler.ssr.get_bundler", return_value=bundler):
            with mock.patch("requests.post", side_effect=mocked_post):
                result = renderer.process("PLACEHOLDER", global_context={})
                self.assertEqual(result, "<div>rendered</div>")

    @override_ap_frontend_settings()
    def test_process_handles_cancelled_response_shape(self):
        bundler = self.create_bundler()
        renderer = BundlerAssetServerSideRenderer({"PLACEHOLDER": DummySSRItem()})

        with mock.patch("alliance_platform.frontend.bundler.ssr.get_bundler", return_value=bundler):
            with mock.patch("requests.post", return_value=MockResponse({"cancelled": True})):
                result = renderer.process("PLACEHOLDER", global_context={})
                self.assertEqual(result, "<!-- SSR_FAILED -->")

    @override_ap_frontend_settings(SSR_CANCEL_ON_TIMEOUT=True)
    def test_timeout_triggers_cancel_request(self):
        bundler = self.create_bundler()
        renderer = BundlerAssetServerSideRenderer({})
        payload = {"requestId": "req-1", "version": 2, "items": {}, "requiredImports": {}}
        called_urls: list[str] = []

        def mocked_post(url, *args, **kwargs):
            called_urls.append(url)
            if url.endswith("/ssr"):
                raise requests.exceptions.Timeout()
            return MockResponse({"cancelled": True})

        with mock.patch("alliance_platform.frontend.bundler.ssr.get_bundler", return_value=bundler):
            with mock.patch("alliance_platform.frontend.bundler.ssr.requests.post", side_effect=mocked_post):
                self.assertIsNone(renderer.process_ssr(payload))

        self.assertTrue(any(url.endswith("/ssr") for url in called_urls))
        self.assertTrue(any(url.endswith("/ssr/cancel") for url in called_urls))



from contextlib import redirect_stdout
import io
import json
from pathlib import Path
import sys
from typing import Any
from typing import Collection

from django.core.management import BaseCommand
from django.core.management.base import OutputWrapper

from .extract_frontend_resources import extract_resources_from_templates

frontend_templates_dir = Path(__file__).parent.parent.parent / "templates"


def extract_asset_filenames() -> tuple[list[Any], dict[str, Collection[str]], list[str], list[str]]:
    """Scans all template files for assets that need to be bundled

    Returns a 4-tuple: list of filenames, breakdown dict, errors and warnings
    """
    resources, breakdown, errors, warnings = extract_resources_from_templates()
    return list(sorted({resource["path"] for resource in resources})), breakdown, errors, warnings


class Command(BaseCommand):
    """
    Extracts used frontend assets from templates used

    This works with any template nodes that extend :class:`~alliance_platform.frontend.bundler.context.BundlerAsset`. All templates
    in the system are loaded to gather all used assets. You can exclude specific directories by setting
    :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.EXTRACT_ASSETS_EXCLUDE_DIRS`
    to either a :class:`pathlib.Path` or ``re.Pattern``.

    Outputs a valid JSON dump as an array of string paths to files.

    Usage::

        # Will write data to output.json
        ./manage.py extract_frontend_assets --output output.json

        # Will write to stdin and supress other std output
        ./manage extract_frontend_assets --quiet
    """

    help = "Outputs JSON dump of files that need to be bundled for production build. This is deprecated in favour of `extract_frontend_resources`."

    def add_arguments(self, parser):
        parser.add_argument(
            "--quiet", action="store_true", help="If set any stdout output will be suppressed"
        )
        parser.add_argument(
            "--output",
            help="The path to write the output to. If not set output will be written to stdout",
        )

    def handle(self, *args, quiet=False, output=None, **kwargs):
        stderr_warn = OutputWrapper(sys.stderr)
        stderr_warn.style_func = self.style.WARNING
        stderr_warn.write(
            "`extract_frontend_assets` is deprecated. `extract_frontend_resources` should be used instead, but requires changes to your Vite config - see the upgrade guide."
        )
        f = io.StringIO()
        with redirect_stdout(f if quiet else self.stdout):
            files, breakdown, errors, warnings = extract_asset_filenames()
            if warnings:
                stderr_warn.write("\n".join(warnings))
            if errors:
                self.stderr.write("\n".join(errors))
            if errors:
                sys.exit(1)
            data = json.dumps({"files": files, "breakdown": breakdown})
            if output:
                Path(output).write_text(data)
            else:
                self.stdout.write(data)



from alliance_platform.frontend.bundler.context import BundlerAssetContext
from alliance_platform.frontend.templatetags.react import CommonComponentSource
from alliance_platform.frontend.templatetags.react import ComponentNode
from alliance_platform.frontend.util import transform_attribute_names
from django.template import Context
from django.template import Origin
from django.test import SimpleTestCase

from tests.test_utils.bundler import bypass_frontend_resource_registry


class TestTransformAttributeNames(SimpleTestCase):
    def test_style_string_is_converted_to_style_object(self):
        attrs = transform_attribute_names({"style": "margin-right: 5px; color: red;"})
        self.assertEqual(
            attrs,
            {
                "style": {
                    "marginRight": "5px",
                    "color": "red",
                }
            },
        )

    def test_style_parser_handles_semicolons_inside_values(self):
        attrs = transform_attribute_names(
            {"style": 'background-image: url("data:image/svg+xml;base64,abc;def"); margin-right: 5px'}
        )
        self.assertEqual(
            attrs["style"],
            {
                "backgroundImage": 'url("data:image/svg+xml;base64,abc;def")',
                "marginRight": "5px",
            },
        )

    def test_style_parser_handles_custom_properties_and_vendor_prefixes(self):
        attrs = transform_attribute_names(
            {"style": "-webkit-line-clamp: 2; -ms-transition: all 1s; --custom-color: red;"}
        )
        self.assertEqual(
            attrs["style"],
            {
                "WebkitLineClamp": "2",
                "msTransition": "all 1s",
                "--custom-color": "red",
            },
        )

    def test_style_parser_preserves_values_as_plain_strings(self):
        attrs = transform_attribute_names(
            {"style": "background-image: url('</script><script>alert(1)</script>')"}
        )
        self.assertEqual(
            attrs["style"],
            {
                "backgroundImage": "url('</script><script>alert(1)</script>')",
            },
        )

    def test_style_parser_ignores_invalid_declarations(self):
        attrs = transform_attribute_names({"style": "margin-right 5px; color: red; : bad; font-size: 12px"})
        self.assertEqual(
            attrs["style"],
            {
                "color": "red",
                "fontSize": "12px",
            },
        )

    def test_style_parser_returns_empty_for_garbage(self):
        attrs = transform_attribute_names({"style": ";;;; not-a-declaration"})
        self.assertEqual(attrs["style"], {})

    def test_style_parser_ignores_unmatched_quotes(self):
        attrs = transform_attribute_names({"style": "color: 'red; font-size: 12px"})
        self.assertEqual(attrs["style"], {})

    def test_style_parser_ignores_unmatched_open_parenthesis(self):
        attrs = transform_attribute_names({"style": "color: red; transform: translate(10px"})
        self.assertEqual(attrs["style"], {"color": "red"})

    def test_style_parser_ignores_unmatched_close_parenthesis(self):
        attrs = transform_attribute_names({"style": "color: red; transform: translateX(10px)); width: 5px"})
        self.assertEqual(
            attrs["style"],
            {
                "color": "red",
                "width": "5px",
            },
        )

    def test_style_parser_ignores_trailing_colon(self):
        attrs = transform_attribute_names({"style": "color: red; width:"})
        self.assertEqual(attrs["style"], {"color": "red", "width": ""})

    def test_style_parser_handles_only_custom_properties(self):
        attrs = transform_attribute_names({"style": "--primary-color: blue; --spacing: 4px;"})
        self.assertEqual(attrs["style"], {"--primary-color": "blue", "--spacing": "4px"})

    def test_component_node_converts_style_prop_string(self):
        with BundlerAssetContext(
            frontend_resource_registry=bypass_frontend_resource_registry, skip_checks=True
        ):
            node = ComponentNode(
                Origin("test"),
                CommonComponentSource("div"),
                {"style": "margin-right: 5px"},
            )
            props = node.resolve_props(Context())
            self.assertEqual(props.props["style"], {"marginRight": "5px"})



import logging
from pathlib import Path
import warnings

from alliance_platform.frontend.bundler.frontend_resource import FrontendResource
from django.conf import settings

logger = logging.getLogger("alliance_platform.frontend")


class FrontendResourceRegistry:
    """Stores any extra resources in addition to those found in templates that should be included in the frontend build

    This should be populated on startup such that the resources are available when
    :djmanage:`extract_frontend_resources` runs.

    The registry to use is specified by the  :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.FRONTEND_RESOURCE_REGISTRY`
    setting.

    These resources are only used when building the frontend in dev. In production, the underlying files do not need to exist -
    for example if your deployment process excludes the source files.

    .. warning::

        In most cases you do not need this.

        This is necessary only for resources that can't be auto-discovered - for example in a dynamic template. If you try
        and use a resource that cannot be auto-discovered an error will be raised.

    Usage::

        frontend_resource_registry = FrontendResourceRegistry()

        frontend_resource_registry.add_resource(
            // Adding a path will be converted to a resource for you based on the extension
            settings.PROJECT_DIR / "frontend/src/file1.tsx",
            // Pass resource directly to have full control
            ESModuleResource(settings.PROJECT_DIR / "frontend/src/file2.tsx", "MyComponent", True),
        )

    """

    _resources: set[FrontendResource]

    def __init__(self):
        self._resources = set()
        self._locked = False

    def add_asset(self, *filenames: Path):
        warnings.warn(
            "`add_asset` is deprecated; call `add_resource` instead", DeprecationWarning, stacklevel=2
        )
        return self.add_resource(*[FrontendResource.from_path(fn) for fn in filenames])

    def add_resource(self, *resources: FrontendResource | Path):
        """Add resource to be included in frontend build"""
        if self._locked:
            raise ValueError(
                "Cannot add resources to registry after it's locked. Make sure all resources are added at startup."
            )
        for resource in resources:
            if isinstance(resource, Path):
                resource = FrontendResource.from_path(resource)
            if not isinstance(resource, FrontendResource):
                raise ValueError(f"Expected FrontendResource but got {type(resource).__name__}")
            self._resources.add(resource)

    def get_resources_for_bundling(self) -> set[FrontendResource]:
        """Get all resources to include in build"""
        return set(self._resources)

    def get_unknown(self, *resources: FrontendResource) -> list[FrontendResource]:
        """From the specified resource(s) return any that aren't in the registry"""
        missing = []
        for resource in resources:
            found = False
            for existing in self._resources:
                if existing.is_substitutable_for(resource):
                    found = True
                    break
            if not found:
                missing.append(resource)
        return missing

    def lock(self):
        self._locked = True
        if settings.DEBUG:
            for resource in self._resources:
                if not resource.path.is_absolute():
                    logging.warning(
                        f"Resources passed to `FrontendResourceRegistry.add_resource` should have absolute paths, received {resource.path}"
                    )
                if not resource.path.exists():
                    logging.warning(
                        f"{resource.path} was added to frontend resource registry but does not exist"
                    )
